﻿:Namespace KeyPress ⍝ V1.07
⍝ 2017 03 02 MBaas: Initial code for KP as UCMD
⍝ 2017 03 09 MBaas: Runs in classic, using 4 digits for hex-codes in non-classic, easy BCol-Fiddling
⍝ 2017 03 10 MBaas: Adjustments to positioning, colour and fonts
⍝ 2017 03 15 Adam: Shorten printed labels, add spacing, adjust fonts, spin off thread if called from session, help text fixes
⍝ 2017 03 16 Adam: Fix help level ?? → ?, remove outputted newline, fix RIU issue
⍝ 2017 03 22 Adam: Made colours consistent

    BCol_Form←¯16   ⍝ = Button Face      188 188 188      ⍝ background of the form
    BCol_Label←¯16  ⍝ = Button Face      166 166 166     ⍝ labels for the various fields
    BCol_Result←¯21 ⍝ = Button Highlight 208 208 208    ⍝ interpreted elements of the message
    BCol_RawMsg←¯21 ⍝ = Button Highlight 208 208 208    ⍝ the event message    (original col was 127 255 255)

    DESC←'Display message arguments of KeyPress event'

    ⎕IO←1 ⋄ ⎕ML←1
    None←{''≡⍵:'[none]' ⋄ ⍵}

    ∇ r←List
      :If ⎕SE.SALTUtils.WIN
          r←⎕NS ⍬
          r.Name←'KeyPress'
          r.Group←'MSWIN'
          r.Desc←DESC
          r.Parse←''
      :Else
          r←⍬
      :EndIf
    ∇

    ∇ r←level Help Cmd;h
      r←⊂DESC
      :If 0<level
          h←'Displays GUI and waits for KeyPress events. '
          h,←'Whenever the event is triggered (i.e. a key is pressed), it displays the arguments. '
          h,←'If run from the session, this data will also be displayed in the session. '
          r,←h''
          h←'More information about the KeyPress-Event: http://help.dyalog.com/'
          h,←1↓∊2↑'.'(=⊂⊢)'.',2⊃⎕SE.SALTUtils.APLV
          h,←'/Content/GUI/MethodOrEvents/KeyPress.htm'
          r,←⊂h
      :Else
          r,←⊂']??',Cmd,' ⍝ for more info'
      :EndIf
    ∇

    ∇ R←Run(Cmd Input)
      R←⍬⊤⍬
      :If ##.RIU
          R←##.RIU Main ⍬
      :Else
          ##.RIU Main&⍬
      :EndIf
    ∇

    ∇ {R}←riu Main dummy;C;E;H;HP;HZ;MSG;P;Q;VP;VZ;W;X;Z;JR;object;event;input;char;key;shift;kl;sz;ShiftS;r
⍝ 03 Apr 1995  Rex Swain, Independent Consultant, Tel (+1) 203-868-0131
⍝ 15 Aug 1995  Simplify and improve handling of System and APL fonts

⍝ ----- Create form ---------------------------------------------------
     
      Q←('Coord' 'Pixel')('Event'('KeyPress' 'Close')1)
      Q,←('MinButton' 0)('MaxButton' 0)('Sizeable' 0)
      'MSG'⎕WC'Form' 'KeyPress Event',Q,⊂'BCol'BCol_Form   ⍝⍎ MSG←
     
     
⍝ ----- Query System and APL fonts ------------------------------------
     
     
      'MSG.SYS'⎕WC ⎕SE.SALTUtils.Fonts.Message ⍝ Create copy of font as object
      'MSG.SYS'⎕WS('Size' 13)('Weight' 800)
      Q←'⎕SE'⎕WG'Font'                     ⍝ Want session/APL font for Labels
      'MSG.APL'⎕WC(⊂'Font'),Q              ⍝ Create copy of font as object
     
      Z←'MSG'⎕WG⊂'TextSize' 'Z' 'MSG.SYS'  ⍝ Size of one char in System font
      C←'MSG'⎕WG⊂'TextSize' '⎕' 'MSG.APL'  ⍝ Size of one char in APL font
     
      H←C[1]⌈Z[1]                          ⍝ Larger height (×1.5 for border)
      W←C[2]                               ⍝ Width of one APL character
      VP←H×¯1.5+2×⍳7                       ⍝ Vertical positions
     
⍝ ----- Resize form ---------------------------------------------------
     
      C←1⊃'.'⎕WG'DevCaps'                ⍝ Screen size in pixels
      Z←(7⊃VP)(2⊃C)                      ⍝ For now, unnecessarily wide
      'MSG'⎕WS('Font' 'MSG.SYS')('Size'Z)
     
⍝ ----- Create column 1: left-hand labels -----------------------------
     
      HP←W                               ⍝ Horizontal position
      VZ←H×1.5                           ⍝ Vertical size
      HZ←⍬                               ⍝ Horizontal size
     
      Q←('Size'(VZ HZ))('BCol'BCol_Label)
      'MSG.L1'⎕WC'Label' ' [1] Object:'('Posn'(1⊃VP)HP),Q
      'MSG.L2'⎕WC'Label' ' [2] Event:'('Posn'(2⊃VP)HP),Q
      'MSG.L3'⎕WC'Label' ' [3] Input Code:'('Posn'(3⊃VP)HP),Q
      'MSG.L4'⎕WC'Label' ' [4] Char Code:'('Posn'(4⊃VP)HP),Q
      'MSG.L5'⎕WC'Label' ' [5] Key Number:'('Posn'(5⊃VP)HP),Q
      'MSG.L6'⎕WC'Label' ' [6] Shift State:'('Posn'(6⊃VP)HP),Q
     
      X←'Label'⎕WN'MSG'                  ⍝ Names of all child labels
      sz←2⊃¨¨X ⎕WG¨⊂'Posn' 'Size'         ⍝ Horizontal posn and size of each
      X ⎕WS¨⊂'Size'⍬(⌈/2⊃¨sz)
      sz←⌈/+/¨sz                           ⍝ Largest Posn+Size
     
     
⍝ ----- Create column 2: APL "edit" boxes -----------------------------
     
      HP←sz                               ⍝ Horiz positions (column 2)
      Q←' KeyPress' ' Ctrl+Shift+Enter'
      HZ←W×2+(↑∘⍴∘,)¨Q
      Q←('Font' 'MSG.APL')('Border' 0)('BCol'BCol_RawMsg)
      JR←⊂'Justify' 'Right'
      'MSG.K1'⎕WC'Label'('Posn'((1⊃VP)HP))('Size'(VZ,1⊃HZ)),Q
      'MSG.K2'⎕WC'Label'('Posn'((2⊃VP)HP))('Size'(VZ,1⊃HZ)),Q
      'MSG.K3'⎕WC'Label'('Posn'((3⊃VP)HP))('Size'(VZ,1⊃HZ)),Q
      'MSG.K4'⎕WC'Label'('Posn'((4⊃VP)HP))('Size'(VZ,1⊃HZ)),Q
      'MSG.K5'⎕WC'Label'('Posn'((5⊃VP)HP))('Size'(VZ,1⊃HZ)),Q
      'MSG.K6'⎕WC'Label'('Posn'((6⊃VP)HP))('Size'(VZ,1⊃HZ)),Q
      Q←(¯1↓Q),⊂'BCol'BCol_Result
     
⍝ ----- Explain input code --------------------------------------------
      E←('Caption' ' ≡')('Font' 'MSG.APL')('Size'VZ(W×2.5))('BCol'BCol_Label)
      :If ⎕SE.SALTUtils.CLASSIC
          P←(3⊃VP),HP+1⊃HZ⍝+W×0.5
          'MSG.Y1'⎕WC'Label'('Posn'P),E
          Z←'MSG.Y1'⎕WG'Size'
          P[2]+←Z[2]
          'MSG.K8'⎕WC'Label'('Posn'P)('Size'(VZ,2⊃HZ)),Q
      :EndIf
     
⍝ ----- Create more ASCII stuff ---------------------------------------
     
      P←(4⊃VP),HP+(1⊃HZ)+W×0.5
      X←('Size'VZ ⍬)('BCol'BCol_Label)
      'MSG.D1'⎕WC'Label' ' decimal'('Posn'P),X
      Z←'MSG.D1'⎕WG'Size'
      P[2]+←Z[2]
      'MSG.D3'⎕WC'Label'('Posn'P),E,⊂'BCol'BCol_Label
      Z←'MSG.D3'⎕WG'Size'
      P[2]+←Z[2]
      'MSG.K7'⎕WC'Label'('Posn'P)('Size'(VZ,0.5×2⊃HZ)),Q
      P[2]+←(0.5×2⊃HZ)
      'MSG.D2'⎕WC'Label' ' hex'('Posn'P),X
     
⍝ ----- Create shift state buttons ------------------------------------
     
      P←(6⊃VP),HP+(1⊃HZ)
      'MSG.B0'⎕WC'Label'('Posn'P),E
      Z←'MSG.B0'⎕WG'Size'
      P[2]+←Z[2]
      'MSG.KS'⎕WC'Label'('Posn'P)('Size'(VZ,2⊃HZ))('BCol'BCol_Result)
      :If 0
          Q←('Style' 'Check')('ReadOnly' 1)('Event' 'Select' ¯1),X
          'MSG.B1'⎕WC'Button' 'Shift'('Posn'P),Q
          Z←'MSG.B1'⎕WG'Size'
          P[2]+←Z[2]+W×1.5
          'MSG.B2'⎕WC'Button' 'Ctrl'('Posn'P),Q
          Z←'MSG.B2'⎕WG'Size'
          P[2]+←Z[2]+W×1.5
          'MSG.B3'⎕WC'Button' 'Alt'('Posn'P),Q
      :EndIf
⍝ ----- Resize width of form ------------------------------------------
     
      X←(⎕WN'MSG')~'MSG.APL' 'MSG.SYS'  ⍝ Names of all children except fonts
      X←2⊃¨¨X ⎕WG¨⊂'Posn' 'Size'    ⍝ Horizontal posn and size of each
      X←W+⌈/+/¨X                    ⍝ Largest Posn+Size, plus right-hand pad
      P←⌈0.30000000000000004 0.9×C-(7⊃VP),X
      'MSG'⎕WS('Size'⍬ X)('Posn'P)  ⍝ Fix overall width and position
     
⍝ ----- Trace keystrokes until Close ----------------------------------
     
      H←⎕D,6⍴⎕A                         ⍝ Hex digits
      Z←⌽'MSG.B1' 'MSG.B2' 'MSG.B3'     ⍝ Shift state buttons
      R←⍬
      :Repeat ⍝ Wait for an event
          X←⎕DQ'MSG'                      ⍝ Wait for event
          :If X[2]≢⊂'Close'
              object event input char key shift←X
              kl←⎕KL input
              'MSG.K1'⎕WS'Caption'(' ',object)
              'MSG.K2'⎕WS'Caption'(' ',event)
              'MSG.K3'⎕WS'Caption'(' ',None input)
              :If ⎕SE.SALTUtils.CLASSIC
                  'MSG.K8'⎕WS'Caption'(' ',kl)
              :EndIf
              'MSG.K4'⎕WS'Caption'(' ',⍕char)
              'MSG.K7'⎕WS'Caption'(' ',H[1+((4-2×⎕SE.SALTUtils.CLASSIC)⍴16)⊤char])
              'MSG.K5'⎕WS'Caption'(' ',⍕key)
              'MSG.K6'⎕WS'Caption'(' ',None⍕shift)
              ⍝Z ⎕WS¨(⊂⊂'State'),¨2 2 2⊤shift
              'MSG.KS'⎕WS'caption'(' ',ShiftS←None ¯1↓∊((2 2 2⊤shift)/'Alt' 'Ctrl' 'Shift'),¨'+')
              :If riu
                  R,←⊂input(⎕SE.SALTUtils.CLASSIC/kl)char key(2 2 2⊤shift)
              :Else
                  r←'Input: ',None input,⎕SE.SALTUtils.CLASSIC/'[',kl,']'
                  r,←', Char: ',⍕char
                  r,←', Key: ',⍕key
                  r,←', Shift: ',(⍕shift),'=',ShiftS ⍝ 1 0⍕2 2 2⊤shift
                  ⎕←r
              :EndIf
          :EndIf
      :Until X[2]≡⊂'Close' ⍝ loop forever
    ∇
:EndNamespace
