﻿:Namespace outputspec ⍝ V1.26
⍝ System user Command
⍝ requires SALTUtils V2.608
⍝
⍝ 2015 05 21 Adam: NS header
⍝ 2016 05 29 DanB: added cmd LastResult and modified box to include a button
⍝ 2016 09 03 DanB: prevent VALERR line 7
⍝ 2016 10 28 DanB: changed SD to reflect better screen dimensions
⍝ 2017 02 21 Adam: Proper ]Box ToolButton and ]LastResult -pfkey → -button
⍝ 2017 02 22 Adam: 2 × -button → -pfkey=, ]?Box overhaul, install button at 1st call and keep it, ]?LastResult overhaul
⍝ 2017 02 23 Adam: Box added new fns and ops for 16.0 and ⎕Uhhhh, added hint and tip to button, do not check for button on non-win, help updates
⍝ 2017 03 02 Adam: disabled ]lastresult altogether and ]box -pfkey
⍝ 2017 03 06 Adam: fix missing trees
⍝ 2017 03 19 Adam: fix ]???rows
⍝ 2017 03 27 Adam: Silence ]rows
⍝ 2017 05 16 JohnS: ]box remove -chars, do not switch on when just changing settings, implement -t=def
⍝ 2017 05 26 Adam: ]???rows note about when cutting occurs
⍝ 2017 06 17 Adam: ]fn.findoutput → ]output.find
⍝ 2017 07 11 Adam: ]?display symbol legend

    ⎕io← ⎕ml←1
    OUTSpace ← '⎕se.Dyalog.Out' ⋄ Var←{0::⍵ ⋄ ⎕se.SALTUtils.LastResultVarName}⍬⊤⍬
    enableLastResult←0
    enableBoxPfkey←0
    enableBoxButton←0

    ∇ r←List;cmds           ⍝ Name, group, short description and parsing rules
      cmds←'LastResult' 'Box' 'Boxing' 'Rows' 'Find'
      r←{⎕NS''}¨cmds        ⍝ space for each cmd.
      r.Name←cmds
      r.Group←⊂'Output'
     
      r[cmds⍳⊂'LastResult'].Desc←⊂'Record Last session output'
      r[cmds⍳'Box' 'Boxing'].Desc←⊂'Diagram display of output'
      r[cmds⍳⊂'Rows'].Desc←'Control of row cutting and folding'
      r[cmds⍳⊂'Find'].Desc←'Find where output is being produced by code'
     
      r[cmds⍳⊂'LastResult'].Parse←'1S  -pfkey='
      r[cmds⍳'Box' 'Boxing'].Parse←⊂'1S -fns="off on" -style="min mid max" -trains="box tree parens def"',enableBoxPfkey/' -pfkey='
      r[cmds⍳⊂'Rows'].Parse←'1S -fns="off on" -style="long cut wrap" -dots= -fold='
      r[cmds⍳⊂'Find'].Parse←'1 -includequadoutput'
     
      r/⍨←enableLastResult∨cmds≢¨⊂'LastResult'
    ∇

    ∇ r←Run(Cmd Input);arg;out;on;sp;cb;O;U;subs;num;PFkey;cap;prev;band;obj
    ⍝ Setup a space for the callback program to run. If it does exist no new space will be created.
      O←⍎OUTSpace ⎕NS'Filter' 'Box' 'Rows' 'SD' 'Dft' 'pfnops' 'flipBox' ⍝'BoxButtonCaptions'
      U←⎕SE.Dyalog.Utils                          ⍝ for lcase, disp, etc.
      sp←'SessionPrint' ⋄ cb←OUTSpace,'.Filter'   ⍝ callbacks
      r←'Was'                                     ⍝ fix for [12109]
     
      :If 0∊⎕NC↑subs←'O.B' 'O.R' 'O.L' 'O.F'      ⍝ spaces for Box, Rows, LastResult and Find
          ⎕NS∘''¨subs                             ⍝ create state-spaces.
          O.(B R).(state fns)←⊂⊂'off'             ⍝ all states OFF
          O.B.(style trains)←'min' 'box'          ⍝ default args for Boxing.
          ⍝O.B.(style trains chars)←'min' 'box' 'regular'  ⍝ default args for Boxing.
          O.R.(style fold dots)←'long' 'off' '·'  ⍝ default args for Rows.
          O.L.state←O.F.state←'off'               ⍝ don't show where output is produced in programs by default
          O.F.includequadoutput←0                 ⍝ don't show ⎕ output by default
          O.L.PFKey←0                             ⍝ no PF key associated with last result
      :EndIf
     
      :Select Cmd
      :CaseList 'Box' 'Boxing'
     
          :If 0=⎕NC'O.B.trains'                   ⍝ Update Tech Preview session
              O.B.trains←'box'                    ⍝ default to boxed fns
          :EndIf
     
          arg←U.lcase⊃Input.Arguments,⊂''         ⍝ single optional argument.
     
          :If enableBoxButton
          :AndIf ⎕SE.SALTUtils.WIN
          :AndIf 0=⎕SE.⎕NC Button ⍝ Add button to toolbar if missing (old BUILDSE)
              Button ⎕WC'ToolButton' 'Boxing'('Hint' 'Toggle boxed display')('Tip' 'Boxing on/off')('Style' 'Check')('State'(O.B.state≡'on'))('ImageIndex' 12)('Event' 'Select'(OUTSpace,'.flipBox'))
              ⎕←'Toolbar button added. Save your session to make permanent.'
          :EndIf
     
          :Select arg
          :CaseList 'off' 'on'
              r←'Was ',U.ucase O.B.state
              O.B.state←arg
              :If enableBoxButton
              :AndIf ⎕SE.SALTUtils.WIN
              :AndIf 0≠⎕NC Button
                  Button ⎕WS'State'(O.B.state≡'on')
              :EndIf
          :Case ,'?'
              r←']boxing ',U.ucase O.B.state
              r,←' -style=',O.B.style
              r,←' -trains=',O.B.trains
              r,←' -fns=',O.B.fns
              ⍝r,←' -chars=',O.B.chars
          :Case ''
              :If 0∧.≡Input.(style trains fns)
              ⍝:If 0∧.≡Input.(style trains chars fns)
                  r←'Is ',U.ucase O.B.state
              :EndIf
          :Case 'reset'
              O.B.(style trains fns)←'min' 'box' 'off'
              ⍝O.B.(style trains fns chars)←'min' 'box' 'off' 'regular'
              r←']boxing ',U.ucase O.B.state
              r,←' -style=',O.B.style
              r,←' -trains=',O.B.trains
              r,←' -fns=',O.B.fns
              ⍝r,←' -chars=',O.B.chars
          :Else
              r←'Arguments: ON OFF RESET ?'
          :EndSelect
     
          :If Input.style≢0                 ⍝ -style=min mid max
              r,←' -style=',O.B.style
              O.B.style←Input.style
              ⍝O.B.state←⊃(arg≡'off')⌽'on' 'off'
          :EndIf
     
          :If Input.trains≢0                ⍝ -trains=box tree parens
              r,←' -trains=',O.B.trains
              O.B.trains←Input.trains
              ⍝O.B.state←⊃(arg≡'off')⌽'on' 'off'
          :EndIf
     
⍝          :If Input.chars≢0                 ⍝ -chars=regular ascii
⍝              r,←' -chars=',O.B.chars
⍝              O.B.chars←Input.chars
⍝          :EndIf
     
          :If Input.fns≢0                   ⍝ -fns=off on
              r,←' -fns=',O.B.fns
              O.B.fns←Input.fns
          :EndIf
     
          :If enableBoxPfkey
          :AndIf Input.pfkey≢0
              :If ⎕SE.SALTUtils.WIN
                  PFkey←{∨/⍵∊0 1:⍬ ⋄ 0=n←n×0 12≠.<n←⌊⊢/∊⎕VFI ⍵~ST←'sScCaA':⍬ ⋄ ⊂'Accelerator'((111+n),+/∪1 1 2 2 4 4 0[ST⍳⍵])}Input.Switch'pfkey'
                  Button ⎕WS PFkey
                  ⎕←'Function key has been assigned. Save your session to make permanent.'
              :Else
                  ⎕←'The modifier -pfkey only works under Windows'
              :EndIf
          :EndIf
     
      :Case 'Rows'                          ⍝ row wrapping/cropping.
          :If 0=⎕NC'O.R.style'              ⍝ Update Tech Preview session
              :Select O.R.state
              :CaseList 'cut' 'long' 'wrap'
                  O.R.(style state←state'on')
              :Else
                  O.R.(style state←'long' 'off')
              :EndSelect
          :EndIf
          arg←U.lcase⊃Input.Arguments,⊂''   ⍝ single optional argument.
     
          :Select arg
          :CaseList 'off' 'on'
              r←'Was ',U.ucase O.R.state
              O.R.state←arg
          :Case r←''
              :If 0∧.≡Input.(style fold dots fns)
                  r←'Is ',U.ucase O.R.state
              :EndIf
          :Case ,'?'
              r←']rows ',U.ucase O.R.state
              r,←' -style=',⍕O.R.style
              r,←' -fns=',O.R.fns
              r,←' -dots=',O.R.dots
              r,←' -fold=',⍕O.R.fold
          :Else
              r←'Arguments: ON OFF ?'
          :EndSelect
     
          :If Input.style≢0                 ⍝ -style=long cut wrap
              r,←(3××⍴r)↓'was -style=',O.R.style
              O.R.style←Input.style
              O.R.state←⊃(arg≡'off')⌽'on' 'off'
          :EndIf
     
          :If Input.fold≢0                 ⍝ -fold=off, 0..9
              :If ~(⊂Input.fold)∊,¨⎕D,⊂'off'
                  r←'* Command Execution Failed: fold must be: off, 0 .. 9'
                  :Return
              :End
              r,←(3××⍴r)↓'was -fold=',O.R.fold
              O.R.fold←Input.fold
              O.R.state←⊃(arg≡'off')⌽'on' 'off'
          :EndIf
     
          :If Input.dots≢0                  ⍝ -dots=.
              r,←(3××⍴r)↓'was -dots=',O.R.dots
              O.R.dots←⊃(Input.dots~''''),'·'
          :EndIf
     
          :If Input.fns≢0                   ⍝ -fns=off on
              r,←(3××⍴r)↓'was -fns=',O.R.fns
              O.R.fns←Input.fns
          :EndIf
     
      :Case 'Find'
          arg←U.lcase⊃Input.Arguments,⊂''   ⍝ single optional argument.
          :Select arg
          :CaseList 'on' 'off'
              r←'Was ',O.F.state
              O.F.state←arg ⋄ O.F.includequadoutput←Input.includequadoutput
          :Case 0 1/¨'?'
              r←'Is ',O.F.state
          :Else
              r←'Arguments: ON OFF ?'
          :EndSelect
     
      :Case 'LastResult'
        ⍝ Set ⎕SE.<global var> to contain the last result and possibly recalled with a PFkey
          arg←U.lcase⊃Input.Arguments,⊂''   ⍝ single optional argument.
          :Select arg
          :CaseList 'on' 'off'
              r←'Was ',prev←O.L.state
              {⍵≠0:_←''⎕PFKEY ⍵}O.L.PFKey ⋄ O.L.PFKey←0 ⍝ disable previous PF key in case new one assigned
              :If 'on'≡O.L.state←arg
                  PFkey←{∨/⍵∊0 1:0 ⋄ 0=n←n×0 12≠.<n←⌊⊢/∊⎕VFI ⍵~ST←'sScCaA':0 ⋄ n+12×+/∪1 1 2 2 4 4 0[ST⍳⍵]}
                  :If 0≠num←PFkey Input.pfkey
                      O.L.PFKey←num⊣Var ⎕PFKEY num
                      ⎕←'PFkey has been set. Save your session to make permanent.'
                  :EndIf
                  ⍎(0∊⎕NC Var)/Var,'←⊂''(no result yet)'''
              :Else ⋄ ⎕EX Var
              :EndIf
          :Case 0 1/¨'?'
              r←'Is ',U.ucase O.L.state
          :Else
              r←'Arguments: ON OFF ? modifier -pfkey='
          :EndSelect
     
      :EndSelect
      on←O.(B R F L).state∨.≢⊂'off'             ⍝ any active?
      '⎕se'⎕WS'Event'sp(⊃on↓0 cb)
    ∇

    ∇ r←level Help Cmd;⎕ML;lev2;n;h
      r←⍬
      :Select Cmd
      :CaseList 'Box' 'Boxing'
          r,←⊂'Diagram display of output.'
          :If 0=level
              r,←⊂'Usage: ]',Cmd,' [on|off|reset|?] [-style={min|mid|max}] [-trains={box|tree|parens|def}] [-fns={off|on}]',' [-pfkey=[S][C][A]nn]'/⍨enableBoxPfkey∧⎕SE.SALTUtils.WIN
              r,←⊂']??',Cmd,' ⍝ for more info'
          :Else ⍝ 1≤level
              :If 1=level
                  r,←⊂'Argument is "on" to activate, "off" to disable, "reset" to restore factory settings, "?" to query state.'
              :Else
                  r,←⊂''
                  r,←⊂'Argument may be any of:'
                  r,←⊂'    (none)  query on/off state only'
                  r,←⊂'    on      activates boxing'
                  r,←⊂'    off     disables boxing'
                  r,←⊂'    reset   restores factory settings: -style=min -trains=box -fns=off'⍝ -chars=regular'
                  r,←⊂'    ?       queries the current state including modifiers'
                  r,←⊂''
              :EndIf
              r,←⊂'-style={min|mid|max} specifies boxing level'
              :If 2≤level
                  r,←⊂'    ┌───┬──────┐    ┌→──┬──────┐    ┌→───────────────┐'
                  r,←⊂'    │min│boxing│    │mid│boxing│    │ ┌→──┐ ┌→─────┐ │'
                  r,←⊂'    └───┴──────┘    └──→┴─────→┘    │ │max│ │boxing│ │'
                  r,←⊂'                                    │ └───┘ └──────┘ │'
                  r,←⊂'                                    └∊───────────────┘'
                  r,←⊂''
                  h←'    The mid and max styles use → to indicate the first axis and ↓ to indicate any additional axes. '
                  h,←'⊖ and ⌽ indicate axes of length 0 and the content is then the prototype rather than actual elements.'
                  r,←⊂h
                  r,←⊂'    The max style uses a character on the lower left border to indicate type as follows:'
                  r,←'        ∊ Nested    ~ Numeric    ─ Character    # Namespace    ∇ ⎕OR    + Mixed' ''
              :EndIf
              r,←⊂'-trains={box|tree|parens|def} specifies display style of trains and derived functions'
              :If 2≤level
                  r,←⊂'    Display of +⌿÷≢ with -trains=...'
                  r,←⊂'        box          tree     parens    def '
                  r,←⊂'    ┌─────┬─┬─┐      ┌─┼─┐    (+⌿)÷≢   +⌿÷≢'
                  r,←⊂'    │┌─┬─┐│÷│≢│      ⌿ ÷ ≢'
                  r,←⊂'    ││+│⌿││ │ │    ┌─┘'
                  r,←⊂'    │└─┴─┘│ │ │    +'
                  r,←⊂'    └─────┴─┴─┘'
                  r,←⊂'    Note that -trains=parens uses default form if any component needs multiple lines to display.'
                  r,←⊂''
              :EndIf
              r,←⊂'-fns={off|on} is whether to also diagram output from running functions'
              :If 2≤level
                  r,←⊂'    Display of {⌽⎕←⍵}''hello'' ''world'' with -fns=...'
                  r,←⊂'         off               on'
                  r,←⊂'     hello  world    ┌─────┬─────┐'
                  r,←⊂'    ┌─────┬─────┐    │hello│world│'
                  r,←⊂'    │world│hello│    └─────┴─────┘'
                  r,←⊂'    └─────┴─────┘    ┌─────┬─────┐'
                  r,←⊂'                     │world│hello│'
                  r,←⊂'                     └─────┴─────┘'
                  r,←⊂''
              :EndIf
⍝              r,←⊂'-chars={regular|ascii} selects character set for drawing'
⍝              :If 2≤level
⍝                  r,←⊂'    ┌───────┬─────┐    .-----.-----.'
⍝                  r,←⊂'    │regular│chars│    |ascii|chars|'
⍝                  r,←⊂'    └───────┴─────┘    ''-----''-----'''
⍝              :EndIf
              :If enableBoxPfkey
              :AndIf ⎕SE.SALTUtils.WIN
                  :If 2≤level
                      r,←⊂''
                  :EndIf
                  r,←⊂'-pfkey=[S][C][A]nn assigns a function key to toggle boxing on/off'
                  :If 2≤level
                      r,←⊂'    ]',Cmd,' -pfkey=12   ⍝ assigns to F12'
                      r,←⊂'    ]',Cmd,' -pfkey=CSA6 ⍝ assigns to Ctrl+Shift+Alt+F6'
                  :EndIf
              :EndIf
              :If 1=level
                  r,←⊂''
                  r,←⊂']???',Cmd,' ⍝ for details'
              :EndIf
          :EndIf
     
      :Case 'Rows'
          r←⊂'Rows are cut, wrapped, scrolled or folded to fit the session window'
          :If level≥1
              r,←⊂''
              r,←⊂'    ]rows [arg] [-switches]    ⍝ Cutting or wrapping of session output'
              r,←⊂'     on                        ⍝ enable row-processing'
              r,←⊂'     off                       ⍝ restore native ⎕PW block-wrapping'
              r,←⊂'     ?                         ⍝ query current state'
              r,←⊂'    -style=long                ⍝ rows extended beyond screen width'
              r,←⊂'    -style=cut                 ⍝ rows truncated at screen width'
              r,←⊂'    -style=wrap                ⍝ each row wrapped at screen width'
              r,←⊂'    -fold=0 .. 9               ⍝ number of trailing rows after fold'
              r,←⊂'    -fold=off                  ⍝ no folding; all rows displayed'
              r,←⊂'    -fns=off on                ⍝ format output from running function'
              r,←⊂'    -dots=.                    ⍝ char to use for ...'
              r,←⊂''
              r,←(lev2←level≥2)↓⊂'    ]???rows        ⍝ for more'
          :AndIf lev2
              r,←⊂''
              h←'Row "folding" replaces rows towards the end of multi-row '
              h,←'output with a line of dots so the expression that gener- '
              h,←'ated the output, together with some leading and trailing '
              h,←'rows, remain visible in the session window. Row folding '
              h,←'may be appropriate for session-based demonstrations.'
              r,←h''
              r,←⊂'      ]rows -fold=3'
              r,←⊂'Was off'
              r,←⊂''
              r,←⊂'      ⍳10 4   ⍝ this assumes that the session window is 13 lines high'
              r,←⊂'┌→───┬────┬────┬────┐'
              r,←⊂'↓1 1 │1 2 │1 3 │1 4 │'
              r,←⊂'├~──→┼~──→┼~──→┼~──→┤'
              r,←⊂'│2 1 │2 2 │2 3 │2 4 │'
              r,←⊂'├~──→┼~──→┼~──→┼~──→┤'
              r,←⊂'│3 1 │3 2 │3 3 │3 4 │'
              r,←⊂'├~──→┼~──→┼~──→┼~──→┤'
              r,←⊂'·····················'
              r,←⊂'├~──→┼~──→┼~──→┼~──→┤'
              r,←⊂'│10 1│10 2│10 3│10 4│'
              r,←⊂'└~──→┴~──→┴~──→┴~──→┘'
              r,←⊂''
              h←'By default, only output from expressions typed into the session is'
              h,←'processed. To include output from running functions, set -fns=on.'
              r,←h''
              r,←⊂'NB: folding is incompatible with wrapping. If both are specified, folding is ignored.'
          :EndIf
     
      :Case 'Find'
          r←⊂'Find where programs produce output without ⎕←'
          :If level=0
              r,←⊂''
              r,←⊂'    ]',Cmd,' arg'
              r,←⊂'     on                        ⍝ enable finding which lines produce output'
              r,←⊂'     off                       ⍝ disable process'
              r,←⊂'     ?                         ⍝ query current state'
              r,←'' 'Lines of programs producing output and NOT starting with ⎕← are displayed'
              r,←⊂'in the session preceded by a line showing which program and which line'
              r,←⊂'produced it. To include lines starting with ⎕← use the switch -includequadoutput.'
              r,←'' 'For an example type'('    ]??',Cmd)
          :Else
              r,←'Example:'('     ]',Cmd,'  on  -includequadoutput')
              r,←⊂'    ∇ foo'
              r,←⊂'[1]   ''Line not starting with ⎕'''
              r,←⊂'[2]   ⎕←''Line with ⎕←'''
              r,←⊂'    ∇'
              r,←⊂'    foo'
              r,←⊂'>> Output from #.foo[1]'
              r,←⊂'Line not starting with ⎕'
              r,←⊂'⎕← Output from #.foo[2]'
              r,←⊂'Line with ⎕←'
     
          :EndIf
     
      :Case 'LastResult'
          r←⊂'Automatically store the last printed result in ',Var,' (variable name may change in a future release).'
          :If 1≤level
              r,←⊂']',Cmd,' [on|off|?] ⍝ ? queries current state'
              r,←'' 'Speed up typing the variable name by assigning it to a function key using the modifier -pfkey=[S][C][A]nn where nn is 1 to 12 and optionally preceeded by S for Shift, C for Control and/or A for Alt:'
              r,←⊂'        ]',Cmd,' on -pfkey=S6 ⍝ store the name in Shift+F6'
              r,←⊂'        ?10⍴100' ⋄ ⎕RL←⍬
              r,←⊂'    ',⍕n←?10⍴100
              r,←⊂'        +/',Var,' ⍝ Use Shift+F6 to type this'
              r,←⊂'    ',⍕,+/n
              r,←'' 'Retain settings across sessions by saving the session.'
          :EndIf
          r,←'' 'If the last printed result was a ref, certain workspace actions be may be temporarily blocked (this restriction may disappear in a future release). To unblock, simply enter a non ref expression in the session.'
          :If 1≤level
              r,←⊂'        ⎕NS ⍬'
              r,←⊂'    #.[Namespace]'
              r,←⊂'        )clear'
              r,←⊂'    Cannot perform operation when # is referenced by session namespace.'
              r,←⊂'        0'
              r,←⊂'    0'
              r,←⊂'        )clear'
              r,←⊂'    clear ws'
          :Else
              r,←''(']??',Cmd,' ⍝ for details')
          :EndIf
     
      :EndSelect
    ∇

    ∇ r←flipBox dummy;c;band;obj
      B.state←'offon'↑⍨3-5×c←B.state≢'on' ⍝ determine new state
      band←⊃⌽'b'⎕SE.cbtop.⎕NL ¯9
      obj←'⎕SE.cbtop.',band,'.tb.boxing'
      obj ⎕WS'State'c  ⍝ change button
      :If c
          '⎕se'⎕WS'Event' 'sessionprint' '⎕se.Dyalog.Out.Filter'  ⍝ turn callback on if necessary
      :EndIf
      r←1
    ∇

    ∇ A Filter _;susp;rand;⎕PP            ⍝ various output filters.
      ⎕PP←⍬⍴⎕RSI.⎕PP                      ⍝ current space's precision.
      :Trap 0 1000                        ⍝ catching errors & interrupts.
          susp←2=-/(,⎕STACK)⍳'*' '⎕DQ'    ⍝ execution suspended.
          susp∨←_≡1                       ⍝ simulate suspension?
          A{⍎'rand←⍺⍺' ⋄ ⍺⍺}0             ⍝ naming of fn/var "rand".
          {                               ⍝ display formatted output.
              isa←2 9∨.=⎕NC'rand'         ⍝ operand is array or namespace.
              O U←⎕SE.Dyalog.(Out Utils)  ⍝ output and utils namespaces.
              rb←'off'∘≢¨O.(R B).state    ⍝ rows and boxing states.
              oon←(2<⍴⎕XSI)∧susp<O.F.state≡'on' ⍝ output on?
              _←⍵{⍵:⍎⎕SE.SALTUtils.LastResultVarName,'←⍺' ⋄ 0}isa∧O.L.state≡'on' ⍝ capture result?
              0 0 0≡rb,oon:⎕←⍵            ⍝ all off: no formatting.
              oon:⍵{0::⎕←⍺ ⋄ (fn lc)←⍵ ⋄ lc←1⌈lc
                  O.F.includequadoutput<'⎕'∊qo←'⎕←>>'/⍨2/⍲\2/'⎕←'≡2↑{(+/∧\' '=⍵)↓⍵}(1+lc)⌷⎕CR fn:⎕←⍺
                  ⎕←⍺⊣⎕←qo,' Output from ',fn,1⌽'][',⍕lc
              }3⊃¨⎕XSI ⎕LC
              fmt←{                       ⍝ formatting style.
                  box←{susp ⍺(⍵⌶Box)⍵}    ⍝ regular boxing.
                  isa:1 box ⍵             ⍝ array boxing.
                  simp←(⎕NC⊂'rand')∊3.1 3.2      ⍝ simple non-derv.
                  simp:0 box ⍵            ⍝ box-formatting array.
                  B.trains≡'tree':rand Dft 1     ⍝ derv & tree-formatting.
                  B.trains≡'box':0 box ⍵  ⍝ box-formatting array.
                  lines←⍵{∆←⍺⍺ ⋄ ≢⎕FMT ⎕CR'∆'}0  ⍝ number of lines in display
                  lines>1:⊂⍵              ⍝ multi-line train: give up.
                  ⎕THIS U.drvSrc'rand'   ⍝ parenthesised derv.
              }
              0 1≡rb:⎕←fmt ⍵              ⍝ boxing only:
              long←O.R.style≡'long'       ⍝ unrestricted screen width.
              _ ⎕PW←SD⌈32767×long         ⍝ ⎕pw temp set to screen width.
              1 0≡rb:⎕←{                  ⍝ cropping only:
                  isa:susp Rows ⎕FMT ⍵    ⍝ array: cropped.
                  ⊂⍵                      ⍝ fn: force ' ∇name'
              }⍵
              ⎕←susp Rows ⎕FMT fmt ⍵      ⍝ cropping if larger than window.
          }A                              ⍝ :: ∇ *
      :Else
          :If ⎕EN≥1000
              '...'⎕SIGNAL 100            ⍝ interrupt
          :Else
              ⎕←A                         ⍝ error
          :EndIf
      :EndTrap
    ∇

      Box←{⎕ML ⎕IO←0                      ⍝ Boxed session output.
     
          susp isa←⍺                      ⍝ from fun and array display.
          susp<B.fns≡'off':⍵              ⍝ no boxing from fn output: done.
          simp←⍵{~isa:0 ⋄ 1=≡,⍺⍺}0        ⍝ is an array and is simple.
     
          fn←{                            ⍝ formatted function.
              isa:1 box ⍵                 ⍝ top level is array.
              (⊂,⍵)∊pfns:1/⍵              ⍝ primitive fn.
              isprimop ⍵:1/⍵              ⍝ primitive op.
              cls←⍺ class ⍵               ⍝ class of function ⍵.
              cls∊+3.1 4.1:tfn ⍺          ⍝ tradfn or tradop:
              cls∊+3.2 4.2:ndfn ⍵         ⍝ named dfn or dop:
              cls∊-3.2 4.2:↑⍵             ⍝ unnamed dfn or dop:
              ~⍺ isderv ⍵:1 box ⍵         ⍝ not derived fn: give up.
              '['≡1⊃⍵:⍺{                  ⍝ +/[1 → +/[1]
                  ax←'[',(⍕2⊃⍵),']'       ⍝ formatted axis
                  (⊃⍺)ax fn(⊃⍵)ax         ⍝ axis as monadic operator
              }⍵
              ~'.'≡1⊃⍵:⍺ ∇¨⍵              ⍝ derived fn.
              0::⍺ ∇¨⍵                    ⍝ error:
              ⍙←⍎⊃⍵                       ⍝ naming left operand,
              9≠⎕NC'⍙':⍺ ∇¨⍵              ⍝ not a space ref:
              (2⊃⍺)∇ 2⊃⍵                  ⍝ dropping space-tagging.
          }                               ⍝ :: C[;] ← vr ∇ nr
     
          tfn←{'    ∇ ',(6↓¯2↓⍵),'∇'}     ⍝ nice 1970s-style display.
     
          isderv←{                        ⍝ looks like a derived fn.
              ~(⊂⍴⍵)∊,¨2 3:0              ⍝ not a 2- or 3-vector: no
              isprimop 1⊃⍵:1              ⍝ primitive operator.
              ∧/(¯2↑⍺)isfn¨¯2↑⍵:1         ⍝ fork
              4=⌊|(1⊃⍺)class 1⊃⍵          ⍝ defined operator
          }
     
          isfn←{                          ⍝ is a function?
              (⊂,⍵)∊pfns:1                ⍝ primitive fn:
              isname ⍵:1                  ⍝ for late-binding.
              3=⌊⍺ class ⍵:1              ⍝ class is 3:
              ⍺ isderv ⍵                  ⍝ derived fn.
          }                               ⍝ :: yes ← vr ∇ cr
     
          isname←{                        ⍝ valid name.
              ~ischarvec ⍵:0              ⍝ non-starter.
              0≤⎕NC ⍵                     ⍝ possible.
          }
     
          ischarvec←{(1=⍴⍴⍵)∧(⎕DR ⍵)∊82 80 160 320}  ⍝ char vector.
     
          isprimop←{                         ⍝ is a primitive operator.
              kvrs←⎕UCS(80=⎕DR'')/9000+16 56 60 18 ⍝ key, variant, rank, stencil
              pops←'/\⌿⍀.¨∘⍨&⍣[⌶@'           ⍝ classic primitive ops
              sops←'⎕S' '⎕R' '⎕OPT' '⎕U2338' '⎕U2360' '⎕U2364' '⎕U233A' ⍝ system ops; ⎕= ⎕: ∘¨ ⎕⋄
              (⊂⍵)∊pops,kvrs,sops            ⍝ is a primitive operator.
          }
     
          ndfn←{                          ⍝ named dfn: removal of name←.
              a←⊃⍵                        ⍝ first line.
              x←a⍳'{'                     ⍝ length of 'name←'.
              ~'⍝'∊a:↑(⊂x↓a),1↓⍵          ⍝ no comment: without name.
              c←a⍳'⍝'                     ⍝ position of line[0] comment.
              d←x↓(c↑a),(x⍴' '),c↓a       ⍝ name← removed, comment adjusted.
              ↑(⊂d),1↓⍵                   ⍝ raw dfn matrix rep.
          }
     
          class←{
              ~ischarvec ⍺:¯1                     ⍝ not a char vector.
              ~'∇     ∇'≡7↑¯1⌽⍺:¯1                ⍝ not ⎕vr
              fx←(⎕NS'').{11::¯1 ⋄ ⎕NC⊂⎕FX ⍵}     ⍝ fix in tmp space.
              (cls←fx ⍵)∊3 4∘.+0.1 0.2:cls        ⍝ trad or named dfn or op:
              -fx,⊂('⍙←',⊃⍵),1↓⍵                  ⍝ unnamed dfn or failure.
          }
     
          box←{                                   ⍝ boxing of ⍵
              U B←⎕SE.Dyalog.(Utils Out.B)        ⍝ handy refs.
              style←B.style                       ⍝ min mid max
              smooth←1⍝B.chars≡'regular'            ⍝ box-drawing chars?
              simp∧~style≡'max':⍵                 ⍝ no boxing.
              style≡'min':0 smooth 0 1 ⎕PP U.disp ⍵   ⍝ min: undecorated disp
              style≡'mid':⍺ smooth 0 1 ⎕PP U.disp ⍵   ⍝ mid: ⍺-decorated disp
              ⍺:smooth ⎕PP U.display ⍵            ⍝ max: ⍺: display
              0 smooth 0 1 U.disp ⍵               ⍝ max: undecorated disp.
          }
     
          uni←80=⎕DR''                            ⍝ unicode interpreter
          pf0←'+-×÷⌊⌈|*⍟<≤=≥>≠∨∧⍱⍲!?~○'           ⍝ scalar fns.
          pf1←'⌷/⌿\⍀∊⍴↑↓⍳⊂⊃∩∪⊣⊢⊥⊤,⍒⍋⍉⌽⊖⌹⍕⍎⍪≡≢⍷'   ⍝ other fns.
          pfu←⎕UCS uni/8838 9080                  ⍝ ⊂_ and ⍳_.
          pfns←,¨pf0,pf1,pfu                      ⍝ primitive fns.
     
          ⍙←⍺⍺ ⋄ 0 box(⊃⎕VR'⍙')fn⊃⎕NR'⍙'          ⍝ boxed output.
      }

      Dft←{⎕IO ⎕ML←0                              ⍝ Display of function tree.
     
          trav←{                                  ⍝ traverse, accumulating subtrees.
              ~(,¨2 3)∨.≡⊂⍴⍵:leaf ⍵               ⍝ not a derv or train: done.
              '['≡1⊃⍵:∇(⊃⍵)('[',(⍕2⊃⍵),']')       ⍝ axis: special treatment
              isop 1⊃⍵:{                          ⍝ derived fn:
                  2=⍴⍵:mop ⍵                      ⍝ monadic operator
                  3=⍴⍵:dop ⍵                      ⍝ dyadic operator.
              }∇¨⍵                                ⍝ formatted subtrees.
              isfn⊃⌽⍵:train ∇¨⍵                   ⍝ train:
              leaf ⍵                              ⍝ neither: give up.
          }
     
          train←{                                 ⍝ function train.
              subs←↑mesh/⍵                        ⍝ enmeshed subtrees.
              tops←apts⊃↓subs                     ⍝ anchor points.
              jpts←mid⍣(2=+/tops),tops            ⍝ joining points.
              xvec←{(3×⍵)++\⍵}jpts                ⍝ character index vector.
              deco←(¯2+⍴⍵)⊃' ── ┌┴┐ ' ' ── ┌┼┐ '  ⍝ plumbing chars.
              (xvec⊃¨⊂deco)⍪subs                  ⍝ subtree matrix.
          }
     
          mop←{                                   ⍝ operator with one operand.
              land oper←↓¨⍵                       ⍝ derived function components.
              tab←+/∧\(⊃land)∊' ┌─┐'              ⍝ indentation for top line.
              leg←dog×⌈spread÷2                   ⍝ length of dog-leg.
              pad←(tab+dog+leg)/' '               ⍝ padding for operator.
              topr←pad∘,¨oper                     ⍝ tabbed operator.
              deco←dog⊃' │  ' ' ┌─┘'              ⍝ plumbing chars.
              join←tab 1 leg dog/deco             ⍝ dog-leg or straight join.
              ↑(topr,↓join),land                  ⍝ subtree matrix.
          }
     
          dop←{                                   ⍝ operator with two operands.
              land oper rand←⍵                    ⍝ derived function components.
              subs←land mesh rand                 ⍝ merged subtrees.
              tops←mid apts⊃↓subs                 ⍝ anchor points.
              xvec←{(3×⍵)++\⍵}tops                ⍝ character index vector.
              head←xvec⊃¨⊂' ── ┌┴┐ '              ⍝ forked subtree joiner.
              otab←(head⍳'┴')/' '                 ⍝ operator padding.
              ↑(otab∘,¨↓oper),↓head⍪subs          ⍝ subtree matrix.
          }
     
          mesh←{                                  ⍝ meshed left and right subtrees.
              lbr rbr←⊂[1 2]↑(⌽⍺)⍵                ⍝ sub branches with same no of rows.
              lft rgt←{+/∧\' '=⍵}¨lbr rbr         ⍝ left and rgt adjacent blanks.
              sep←⌊/lft+rgt                       ⍝ narrowest separation.
              dpl←lft⌊sep                         ⍝ chars to drop from left subtree.
              dpr←0⌈sep-dpl                       ⍝   ..      ..      right   ..
              lvec←⌽¨dpl↓¨↓lbr                    ⍝ left rows.
              rvec←dpr↓¨(spread/' ')∘,¨↓rbr       ⍝ right rows, padded with min gap.
              trim←{(~∧\∧⌿⍵=' ')/⍵}               ⍝ trim off outer blank cols.
              ⌽trim⌽trim↑lvec,¨rvec               ⍝ meshed subtrees.
          }
     
          leaf←{                                  ⍝ format leaf.
              ⍬≡⍵:'⍬'                             ⍝ ⍬ is special
              ''≡⍵:⍉⍪''''''                       ⍝ '' is special
              1∊'←{'⍷⊃↓⍵:∇↑noname↓⍵               ⍝ dfn without name.
              dots←{(⍵⍳⍺)⊃¨⊂⍵,'·'}                ⍝ dots for blanks.
              {⍵ dots(,⍵)~' '}⎕FMT ⍵              ⍝ char matrix.
          }
     
          noname←{                                ⍝ without dfn name.
              a←⊃⍵                                ⍝ first line.
              x←a⍳'{'                             ⍝ length of 'name←'.
              ~'⍝'∊a:(⊂x↓a),1↓⍵                   ⍝ no comment: without name.
              c←a⍳'⍝'                             ⍝ position of line[0] comment.
              d←x↓(c↑a),(x⍴' '),c↓a               ⍝ name← removed, comment adjusted.
              (⊂d),1↓⍵
          }
     
          isfn←{                                  ⍝ is function?
              0=≡⍵:⍵∊pfns                         ⍝ primitive function:
              dfnop ⍵:1                           ⍝ dfn:
              ~(,¨2 3)∨.≡⊂⍴⍵:0                    ⍝ not a derv or train.
              isop 1⊃⍵:1                          ⍝ derv.
              ∇⊃⌽⍵                                ⍝ derv or train.
          }
     
          isop←{                                  ⍝ is operator?
              (⊂⍵)∊pops:1                         ⍝ primitive operator:
              ']['≡2↑¯1⌽⍵:1                       ⍝ axis: [1 2 ..]
              ~dfnop ⍵:0                          ⍝ not a d-op:
              1∊↑'⍺⍺' '⍵⍵'⍷¨⊂⍵                    ⍝ is a d-op.
          }
     
          mid←{⍵∨(⍳⍴⍵){⍺=⌊(+/⍵/⍺)÷2}⍵}            ⍝ mask with midpoint
          trim←{⍵-¯1⌽1 1⍷⍵}⍣≡                     ⍝ for ('abc',)
          apts←trim∘(~∘(∊∘' ┌─┐'))                ⍝ anchor points for sub-trees.
          dfnop←{'}'≡⊃⌽~∘' ',⍵}                   ⍝ dfn or dop
     
          pfns pops←pfnops 80=⎕DR''               ⍝ unicode-sensitive list of fns&ops.
     
          spread←1⌈⍵                              ⍝ horizontal spread.
          ⍺←1 ⋄ dog←⍺                             ⍝ dog-leg for monadic operators.
     
          fn←⍺⍺ ⋄ trav ⎕CR'fn'                    ⍝ display of function tree.
      }

      pfnops←{                                    ⍝ primmitive fns and ops.
          pf0←'+-×÷⌊⌈|*⍟<≤=≥>≠∨∧⍱⍲!?~○'           ⍝ scalar fns.
          pf1←'⌷/⌿\⍀∊⍴↑↓⍳⊂⊃∩∪⊣⊢⊥⊤,⍒⍋⍉⌽⊖⌹⍕⍎⍪≡≢⍷'   ⍝ other fns.
          pfu←⎕UCS ⍵/8838 9080                    ⍝ ⊂_ and ⍳_.
          pfns←pf0,pf1,pfu                        ⍝ primitive fns.
     
          kvrs←⎕UCS ⍵/9000+16 56 60 18            ⍝ key, variant, rank, stencil
          pops←'/\⌿⍀.¨∘⍨&⍣[⌶@'                    ⍝ classic primitive ops
          sops←'⎕S' '⎕R' '⎕OPT' '⎕U2338' '⎕U2360' '⎕U2364' '⎕U233A' ⍝ system ops. ⎕= ⎕: ∘¨ ⎕⋄
          ops←pops,kvrs,sops                      ⍝ operators.
          pfns ops                                ⍝ fns & ops.
      }

      Rows←{                               ⍝ Cropped to fit session window.
          R←⎕SE.Dyalog.Out.R               ⍝ ref for params.
          ⍺<R.fns≡'off':⍵                  ⍝ no cropping from fn output: done.
          R.style≡'wrap':{0 0⍴''}{⎕←⍵}¨↓⍵  ⍝ ⎕PW-wrap each line.
          sr sc←SD⌈0,32767×R.style≡'long'  ⍝ window size
          rows cols←⍴⍵                     ⍝ raw output size.
          fold dot←R.(fold dots)           ⍝ params
          coldots←{                        ⍝ with column dots.
              cols≤sc:⍵                    ⍝ no column cropping: done.
              1≠rows:(0 ¯1↓⍵),dot          ⍝ single column of dots.
              (0 ¯3↓⍵),1 3⍴dot             ⍝ single row: stuff ...
          }
          dc←cols⌊sc                       ⍝ number of cols to display.
          nofold←(fold≡'off')∨rows≤sr-2    ⍝ no folding
          nofold:coldots rows dc↑⍵         ⍝ row cropping only.
          tail←⍎fold                       ⍝ 0..9
          t←0⌈tail⌊sr-3                    ⍝ no of trailing rows.
          h←0⌈sr-3+t                       ⍝ input + brk + fold + prompt.
          brk←sc⍴dot                       ⍝ break marker ···
          top←h sc↑⍵                       ⍝ retained upper part of output.
          bot←(t sc×¯1 1)↑⍵                ⍝    ..    lower  ..     ..
          mat←top⍪brk⍪bot                  ⍝ dots-broken matrix
          dr _←⍴mat                        ⍝ number of rows to display.
          coldots dr dc↑mat                ⍝ cropped.
      }

    ∇ rc←SD;handle;GetDlgItem;GetClientRect;GetDC;ReleaseDC;GetDeviceCapsdlg;r;c;v;h;_;scal;hdc  ⍝ session window size
      :Trap 0
          ⎕NA'u user32|GetDlgItem u u'
          ⎕NA'u user32|GetClientRect u >{u u u u}'
          handle←'⎕se'⎕WG'Handle'
          dlg←GetDlgItem handle 0
          _(_ _ c r)←GetClientRect dlg(0 0 0 0)
          scal←1
          v h←2 ⎕NQ'⎕se' 'GetTextSize' 'X'
          :If (,'1')≡2 ⎕NQ'.' 'GetEnvironment' 'AutoDPI'
              ⎕NA'u user32|GetDC U'
              ⎕NA'u user32|ReleaseDC U U'
              ⎕NA'u gdi32|GetDeviceCaps U U'
              scal←96÷⍨GetDeviceCaps(hdc←GetDC 0)88
              {}ReleaseDC 0 hdc
          :End
     
          rc←⌊r c÷v h×scal
      :Else
          rc←⎕SD-1 0  ⍝ Unix version?
      :EndTrap
    ∇

    ∇ obj←Button;band ⍝ name of ]boxing button for ⎕W_
      band←⊃⌽'b'⎕SE.cbtop.⎕NL ¯9
      obj←'⎕SE.cbtop.',band,'.tb.boxing'
    ∇

:EndNamespace

 ⍝ outputspec  $Revision: 30431 $
